# Copyright 2021 Katteli Inc.
# TestFlows.com Open-Source Software Testing Framework (http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# to the end flag
import contextvars

from collections import namedtuple

from concurrent.futures import CancelledError
from concurrent.futures import TimeoutError
from concurrent.futures import Future as ConcurrentFuture

from .asyncio import asyncio, wrap_future, OptionalFuture
from .asyncio import Future as AsyncFuture
from .asyncio import is_running_in_event_loop
from .asyncio import TimeoutError as AsyncTimeoutError
from .asyncio import CancelledError as AsyncCancelledError

def Context(**kwargs):
    """Convenience function to create
    a namedtuple to store parallel context variables.
    """
    return namedtuple("ParallelContext", " ".join(kwargs.keys()))(*kwargs.values())

context = Context(
    current=contextvars.ContextVar('_testflows_current', default=None),
    previous=contextvars.ContextVar('_testflows_previous', default=None),
    top=contextvars.ContextVar('_testflows_top', default=None),
    is_valid=contextvars.ContextVar('_testflows_is_valid', default=None),
)

# set current parallel context as valid
context.is_valid.set(True)

ContextVar = contextvars.ContextVar
copy_context = contextvars.copy_context

def convert_result_to_concurrent_future(fn, args=None, kwargs=None):
    """Make concurrent future out of result of a function call.
    """
    if args is None:
        args = ()
    if kwargs is None:
        kwargs = {}

    future = ConcurrentFuture()
    future.set_running_or_notify_cancel()
    try:
        future.set_result(fn(*args, **kwargs))
    except BaseException as exc:
        future.set_exception(exc)

    return future

def _get_parallel_context():
    """Return parallel context based on contextvars
    with all the user context variables cleared
    to None.
    """
    ctx = contextvars.copy_context()
    # clear any user context variables to None
    for var in ctx.keys():
        if not var.name.startswith("_testflows_"):
            # in Python 3.8 ContextVar can't be cleared
            # and therefore the best we can do is to set
            # user context variable to None
            var.set(None)
    return ctx

def _check_parallel_context():
    """Check if parallel parallel context is valid.
    """
    if not context.is_valid.get():
        raise RuntimeError("parallel context was not set")

def top(value=None):
    """Highest level test.
    """
    if value is not None:
        context.top.set(value)
    return context.top.get()

def current(value=None, set_value=False):
    """Currently executing test.
    """
    if value is not None or set_value:
        context.current.set(value)
    return context.current.get()

def previous(value=None):
    """Last executed test.
    """
    if value is not None:
        context.previous.set(value)
    return context.previous.get()

def join(*future, futures=None, test=None, filter=None, all=False, no_async=False, cancel_pending=False):
    """Wait for parallel test futures to complete.
    Returns a list of completed tests.
    
    Terminates current test if any of the parallel
    tests raise an exception.

    If no futures specified uses test.futures().
    If test is not specified test is set to current().

    :param *future: one or more futures (optional)
    :param futures: list of futures (optional)
    :param test: current test, default: current()
    :param filter: filter function, default: None
    :param all: wait and join all the tests, default: False
    :param no_async: force non async join while running in event loop
    :param cancel_pending: cancel any pending futures
    """
    if no_async is False and is_running_in_event_loop():
        return _async_join(*future, futures=futures, test=test, filter=filter, all=all, cancel_pending=cancel_pending)

    if test is None:
        test = current()

    futures = list(future) or futures or test.futures
    tests = []
    exception = None
    filtered_count = 0

    while True:
        if not futures or filtered_count >= len(futures):
            break

        future = None
        try:
            future = futures.pop(0)

            if filter and not filter(future):
                futures.append(future)
                filtered_count += 1
                continue

            filtered_count = 0

            if cancel_pending:
                if future.cancel():
                    continue
            try:
                exc = future.exception(timeout=0.1)
                if exc is not None:
                    if exception is None:
                        exception = exc
                        if test:
                            test.terminate()
                    if not all:
                        raise exception
                else:
                    tests.append(future.result(timeout=0.1))
            except CancelledError:
                continue
            except TimeoutError:
                futures.append(future)
                continue
        except BaseException:
            if future is not None:
                futures.append(future)
            raise

    if exception is not None:
        raise exception
    return tests

async def _async_join(*future, futures=None, test=None, filter=None, all=False, cancel_pending=False):
    """Wait for async parallel test futures to complete.
    Returns a list of completed tests.
    
    Terminates current test if any of the parallel
    tests raise an exception.

    If no futures specified uses test.futures().
    If test is not specified test is set to current().

    :param *future: one or more futures (optional)
    :param futures: list of futures (optional)
    :param test: current test, default: current()
    :param filter: filter function, default: None
    :param all: wait and join all the tests, default: False
    :param cancel_pending: cancel any pending futures
    """
    if test is None:
        test = current()

    futures = list(future) or futures or test.futures
    tests = []
    exception = None
    filtered_count = 0

    while True:
        if not futures or filtered_count >= len(futures):
            break

        future = None
        try:
            future = futures.pop(0)

            if filter and not filter(future):
                futures.append(future)
                filtered_count += 1
                continue

            filtered_count = 0

            try:
                f = future
                if isinstance(future, ConcurrentFuture):
                    f = wrap_future(future, new_future=OptionalFuture())

                if cancel_pending:
                    if future.cancel():
                        continue

                tests.append(await asyncio.wait_for(asyncio.shield(f), timeout=0.1))
            except AsyncCancelledError:
                continue
            except AsyncTimeoutError:
                futures.append(future)
                continue
        except BaseException as exc:
            if future is not None:
                if future.done():
                    if test:
                        test.terminate()
                    if not all:
                        raise
                    if exception is None:
                        exception = exc
                    continue
                else:
                    futures.append(future)
            raise

    if exception is not None:
        raise exception
    return tests



import sys
import traceback

def exception(exc_type=None, exc_value=None, exc_traceback=None):
    """Get exception string.
    """
    if (exc_type, exc_value, exc_traceback) == (None, None, None):
        exc_type, exc_value, exc_traceback  = sys.exc_info()
    return "".join(traceback.format_exception(exc_type, exc_value, exc_traceback)).rstrip()


class TestFlowsException(Exception):
    """Base exception class.
    """
    pass

class ResultException(TestFlowsException):
    """Result exception.
    """
    pass

class DummyTestException(TestFlowsException):
    """Dummy test exception.
    """
    pass

class TestIteration(TestFlowsException):
    """Repeat test.
    """
    def __init__(self, repeat, retry, *args, **kwargs):
        self.repeat = repeat
        self.retry = retry
        super(TestIteration, self).__init__(*args, **kwargs)

class TestRerunIndividually(TestFlowsException):
    """Repeat tests individually.
    """
    def __init__(self, tests, *args, **kwargs):
        self.tests = tests
        super(TestRerunIndividually, self).__init__(*args, **kwargs)

class TestFlowsError(TestFlowsException):
    """Base error exception class.
    """
    pass

class RequirementError(TestFlowsError):
    """Requirement error.
    """
    pass

class SpecificationError(TestFlowsError):
    """Specification error.
    """
    pass

class DescriptionError(TestFlowsError):
    """Description error.
    """
    pass

class ArgumentError(TestFlowsError):
    """Argument error.
    """
    pass

class TerminatedError(TestFlowsError):
    """Terminated error.
    """
    pass


# Copyright 2019 Katteli Inc.
# TestFlows.com Open-Source Software Testing Framework (http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import io
import sys
import csv
import argparse

from argparse import ArgumentTypeError
from collections import namedtuple
from testflows._core.exceptions import exception
from testflows._core.compress import CompressedFile
from testflows._core.objects import Repeat, Retry
from testflows._core.tracing import logging

import testflows._core.contrib.rsa as rsa

KeyValue = namedtuple("KeyValue", "key value")
NoneValue = "__none__"

class FileType(object):
    def __init__(self, mode='r', bufsize=-1, encoding=None, errors=None, safe=False):
        self._mode = mode
        self._bufsize = bufsize
        self._encoding = encoding
        self._errors = errors
        self._safe = safe

    def __call__(self, string):
        # the special argument "-" means sys.std{in,out}
        if string == '-':
            if 'r' in self._mode:
                if 'b' in self._mode:
                    return sys.stdin.buffer
                return sys.stdin
            elif 'w' in self._mode:
                if 'b' in self._mode:
                    return sys.stdout.buffer
                return sys.stdout
            else:
                msg = argparse._('argument "-" with mode %r') % self._mode
                raise ValueError(msg)

        # all other arguments are used as file names
        try:
            if self._safe and os.path.exists(string):
                raise FileExistsError(f"File '{string}' already exists, please delete it first.")

            return open(string, self._mode, self._bufsize, self._encoding,
                        self._errors)
        except OSError as e:
            message = argparse._("can't open '%s': %s")
            raise ArgumentTypeError(message % (string, e))

    def __repr__(self):
        args = self._mode, self._bufsize
        kwargs = [('encoding', self._encoding), ('errors', self._errors)]
        args_str = ', '.join([repr(arg) for arg in args if arg != -1] +
                             ['%s=%r' % (kw, arg) for kw, arg in kwargs
                              if arg is not None])
        return '%s(%s)' % (type(self).__name__, args_str)

class LogFileType(object):
    def __init__(self, mode='r', bufsize=-1, encoding=None, errors=None):
        self._mode = mode
        self._encoding = encoding
        self._errors = errors

    def __call__(self, string):
        # the special argument "-" means sys.std{in,out}
        if string == '-':
            if 'r' in self._mode:
                fp = CompressedFile(sys.stdin.buffer, self._mode)
                if self._encoding:
                    return io.TextIOWrapper(fp, self._encoding, self._errors)
                return fp
            elif 'w' in self._mode:
                fp = CompressedFile(sys.stdout.buffer, self._mode)
                if self._encoding:
                    return io.TextIOWrapper(fp, self._encoding, self._errors)
                return fp
            else:
                msg = argparse._('argument "-" with mode %r') % self._mode
                raise ValueError(msg)

        # all other arguments are used as file names
        try:
            fp = CompressedFile(string, self._mode)
            if self._encoding:
                return io.TextIOWrapper(fp, self._encoding, self._errors)
            return fp
        except OSError as e:
            message = argparse._("can't open '%s': %s")
            raise ArgumentTypeError(message % (string, e))

    def __repr__(self):
        args = self._mode
        kwargs = [('encoding', self._encoding), ('errors', self._errors)]
        args_str = ', '.join([repr(arg) for arg in args if arg != -1] +
                             ['%s=%r' % (kw, arg) for kw, arg in kwargs
                              if arg is not None])
        return '%s(%s)' % (type(self).__name__, args_str)

def path(p, special=""):
    if p in special or []:
        return p
    if not os.path.exists(os.path.abspath(p)):
        raise ArgumentTypeError(f"path does not exist: '{os.path.abspath(p)}'")
    return p

def file(*args, **kwargs):
    """File type."""
    return FileType(*args, **kwargs)

def logfile(*args, **kwargs):
    """Log file type."""
    return LogFileType(*args, **kwargs)

def rsa_private_key_pem_file(p):
    """RSA private key PEM file type.
    """
    with open(p, mode="rb") as pem_file:
        return rsa.PrivateKey.load_pkcs1(pem_file.read())

def key_value(s, sep='='):
    """Parse a key, value pair using a seperator (default: '=').
    """
    if sep not in s:
        raise ArgumentTypeError(f"invalid format of key{sep}value")
    key, value= s.split(sep, 1)
    return KeyValue(key.strip(), value.strip())

def count(value):
    try:
        value = int(value)
        assert value >= 0
    except:
        raise ArgumentTypeError(f"{value} is not a positive number")
    return value

def repeat(value):
    try:
        fields = list(csv.reader([value], "unix"))[-1]
        until = "fail"
        if len(fields) > 2:
            pattern, count, until = fields
        else:
            pattern, count = fields
        option = Repeat(count=count, pattern=pattern, until=until)
    except Exception as e:
        raise ArgumentTypeError(f"'{value}' is invalid")
    return option

def retry(value):
    try:
        fields = list(csv.reader([value], "unix"))[-1]
        jitter = None # can't be specified
        if len(fields) < 2:
            raise ValueError("needs at least 2 values")
        pattern, count, timeout, delay, backoff = [*fields, *["", None, None, 0, 1][len(fields):]]
        if not pattern:
            pattern = ""
        if not count:
            count = None
        if not timeout:
            timeout = None
        if not delay:
            delay = 0
        if not backoff:
            backoff = 1
        option = Retry(count=count, timeout=timeout, delay=delay, backoff=backoff, jitter=jitter, pattern=pattern)
    except Exception as e:
        raise ArgumentTypeError(f"'{value}' is invalid")
    return option

def tags_filter(value):
    try:
        type, cvstags = value.split(":", 1)
        assert type in ["test", "suite", "module", "feature", "scenario", "any"]
        tags = list(csv.reader([cvstags], "unix"))[-1]
        if type == "scenario":
            type = "test"
        elif type == "feature":
            type = "suite"
        option = (type, tuple(tags))
    except Exception as e:
        raise ArgumentTypeError(f"'{value}' is invalid")
    return option

def onoff(value):
    if value.lower() in ["yes", "1", "on"]:
        return True
    elif value.lower() in ["no", "0", "off"]:
        return False
    elif value == NoneValue:
        return NoneValue
    raise ArgumentTypeError(f"'{value}' is invalid")

onoff.metavar = str(set(["yes", "1", "on", "no", "0", "off"])).replace(" ","").replace("'","")

def trace_level(value):
    if value.lower() in ["debug", "info", "warning", "error", "critical"]:
        return getattr(logging, value.upper())
    else:
        raise ArgumentTypeError(f"'{value}' is invalid")

trace_level.choices = ["debug", "info", "warning", "error", "critical"]
trace_level.metavar = str(set(trace_level.choices)).replace(" ","").replace("'","")


# Copyright 2019 Katteli Inc.
# TestFlows.com Open-Source Software Testing Framework (http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from hashlib import sha1
from collections import namedtuple

InitArgs = namedtuple("InitArgs", "args kwargs")

def namedtuple_with_defaults(*args, defaults=()):
    nt = namedtuple(*args)
    nt.__new__.__defaults__ = defaults
    [setattr(nt, f"_{field}", idx) for idx, field in enumerate(nt._fields)]
    return nt

def get(a, b):
    """a if not a is None else b.
    """
    return a if a is not None else b

def hash(*s, short=False):
    """Calculate standard hash.

    :param s: strings
    :param short: short version, default: False
    """
    value = sha1(''.join(s).encode("utf-8")).hexdigest()[:32]
    if short:
        return value[-16:]
    return value


class TestObject(object):
    """Base class for all the test objects.
    """
    #: object fields used to represent state of the object
    _fields = ()
    #: defaults for the fields
    _defaults = ()

    def __new__(cls, *args, **kwargs):
        obj = super(TestObject, cls).__new__(cls)
        obj.initargs=InitArgs(
            args=[a for a in args],
            kwargs={k: v for k,v in kwargs.items()})
        return obj

    @property
    def id(self):
        return hash(*[repr(getattr(self, field)) for field in self._fields if field != "id"])

    def __iter__(self):
        return iter([getattr(self, field) for field in self._fields])

    def __repr__(self):
        """Custom object representation.
        """
        args = ",".join([repr(arg) for arg in self.initargs.args])
        kwargs = ",".join([name + "=" + repr(value) for name, value in self.initargs.kwargs.items()])
        name = self.__class__.__name__
        if args and kwargs:
            args += ","
        return "%s(%s%s)" % (name, args, kwargs)


class TestArg(TestObject):
    """Base class for all test argument object.
    """
    pass

class Table(tuple, TestObject):
    _fields = ("header", "rows", "row_format")
    _defaults = (None, ) * 3
    _row_type_name = "Row"

    def __new__(cls, header=None, rows=None, row_format=None, _row_type=None):
        if rows is None:
            rows = []
        if header is None:
            header = ""

        if _row_type is None:
            row_type = namedtuple(cls._row_type_name, header)
        else:
            row_type = _row_type

        class Row(row_type):
            def __str__(self):
                s = super(Row, self).__str__()
                return s.split("(", 1)[-1].rstrip(")") or s

            @property
            def __dict__(self):
                return self._asdict()

        obj = super(Table, cls).__new__(cls, [Row(*row) for row in rows])
        obj.initargs=InitArgs(
            args=[header, rows, row_format],
            kwargs={})
        obj.header = header
        obj.rows = obj
        obj.row_type = row_type
        if row_format:
            row_format % tuple(obj.header.split(" "))
        else:
            row_format = Table.default_row_format(row_type._fields, obj[0] if obj else None)
        obj.row_format = row_format
        return obj

    @staticmethod
    def default_row_format(fields, row):
        # if no row_formatter is given then use the header as an example
        # or first row, whichever is longer
        sample_row = fields
        if row and len(str(fields)) < len(str(tuple(row))):
            sample_row = row
        row_format = " | ".join([f"%-{max(len(str(c)), 3)}s" for c in sample_row])
        return row_format

    @staticmethod
    def __str_header__(fields, row_format):
        s = [row_format % fields]
        s.append(row_format % tuple(["-" * len(field) for field in fields]))
        return "\n".join(s)

    @staticmethod
    def __str_row__(row, row_format):
        return row_format % row

    def __str__(self):
        """Return markdown-styled table representation.
        """
        s = [Table.__str_header__(self.row_type._fields, self.row_format)]
        s += [Table.__str_row__(row, self.row_format) for row in self]
        return "\n".join(s)

    @classmethod
    def from_table(cls, table):
        """Creates table from a table.
        """
        return cls(header=" ".join(table.row_type._fields), rows=table, row_format=table.row_format)
